from src.models.request import Request
from src.preprocessing import evaluate


class Step:
    """Class for the Step model"""
    def __init__(
        self,
        step: dict,
        steps: dict,
    ):
        for key in step:
            step[key] = evaluate(step[key], steps)

        self.name = step["name"]
        self.operation_id = step["operation_id"]
        self.method = step["method"]
        self.path = step["path"]
        self.status_code = step["status_code"]

        if "headers" in step:
            self.headers = step["headers"]
        else:
            self.headers = None

        if "body" in step:
            self.body = step["body"]
        else:
            self.body = None

        if "raw_body" in step:
            self.body = step["raw_body"]

        if "params" in step:
            self.params = step["params"]
        else:
            self.params = None

        if "auth" in step:
            self.auth = step["auth"]
        else:
            self.auth = None

    def construct_request(self, base_url, global_auth):
        return Request(
            base_url=base_url,
            method=self.method,
            path=self.path,
            params=self.params,
            headers=self.headers,
            global_auth=global_auth,
            auth=self.auth,
            body=self.body,
        )



import pytest
from hamcrest import assert_that, is_, has_items, instance_of

from src.parsing import *


class TestParsing:
    """Class for basic unit testing of the parsing.py module"""

    steps_file_path = "./tests/fixtures/valid/steps/person_api.yml"
    spec_file_path = "./tests/fixtures/valid/specs/person_api.yml"

    invalid_steps_missing_operation_id_file_path = (
        "./tests/fixtures/invalid/steps/missing_property.yml"
    )
    invalid_spec_missing_openapi_property_file_path = (
        "./tests/fixtures/invalid/specs/invalid_spec_1.yml"
    )

    test_steps = parse_steps_file(steps_file_path)

    def test_parse_steps_file_with_valid_steps(self):
        """Ensure a step file is parsed correctly"""
        steps = parse_steps_file(self.steps_file_path)

        assert_that(steps, is_(instance_of(dict)))
        assert_that(steps, has_items("coverage_threshold", "base_url", "steps"))
        assert_that(steps["coverage_threshold"], is_(instance_of(float)))
        assert_that(steps["base_url"], is_(instance_of(str)))
        assert_that(steps["steps"], is_(instance_of(list)))

        assert_that(
            steps["steps"][0],
            has_items(
                "name",
                "operation_id",
                "method",
                "path",
                "headers",
                "body",
                "auth",
                "status_code",
            ),
        )
        assert_that(steps["steps"][0]["name"], is_(instance_of(str)))
        assert_that(steps["steps"][0]["operation_id"], is_(instance_of(str)))
        assert_that(steps["steps"][0]["method"], is_(instance_of(str)))
        assert_that(steps["steps"][0]["path"], is_(instance_of(str)))
        assert_that(steps["steps"][0]["headers"], is_(instance_of(dict)))
        assert_that(steps["steps"][0]["body"], is_(instance_of(dict)))
        assert_that(steps["steps"][0]["auth"], is_(instance_of(dict)))
        assert_that(steps["steps"][0]["status_code"], is_(instance_of(int)))

        assert_that(
            steps["steps"][1],
            has_items(
                "name",
                "operation_id",
                "method",
                "path",
                "headers",
                "auth",
                "status_code",
            ),
        )

    def test_parse_steps_file_with_missing_operation_id(self):
        """Ensure that a missing required property raises a ValidationError
        ValidationError comes from the jsonschema library
        """
        with pytest.raises(ValidationError) as e:
            invalid_steps = parse_steps_file(
                self.invalid_steps_missing_operation_id_file_path
            )

    def test_parse_valid_spec_file(self):
        """Ensure an OpenAPI document is parsed correctly"""
        spec, operation_schemas = parse_spec_file(self.test_steps, self.spec_file_path)
        assert_that(spec, is_(instance_of(dict)))
        # TODO: More tests to assert children are properly formatted
        assert_that(operation_schemas, is_(instance_of(dict)))
        # TODO: More tests

    def test_parse_invalid_spec_file(self):
        """Ensure an OpenAPI document is parsed correctly
        OpenAPIValidationError comes from openapi_spec_validator library
        """
        with pytest.raises(OpenAPISpecValidatorError):
            spec, operation_schemas = parse_spec_file(
                self.test_steps, self.invalid_spec_missing_openapi_property_file_path
            )



import os
import json

import yaml
from jsonschema import validate, ValidationError
from json_ref_dict import materialize, RefDict
from openapi_spec_validator import validate as validate_spec
from openapi_spec_validator.exceptions import OpenAPISpecValidatorError

from .errors import *


def parse_steps_file(step_file_path: str) -> dict:
    """Parse a .yml steps file and convert it to a python dictionary

    Args:
        step_file_path (str): Relative path to steps file

    Raises:
        FileNotFoundError: Step file not found
        ValidationError: Step file not valid

    Returns:
        dict: Step file converted to a python dictionary
    """
    try:
        steps_schema_path = os.path.join(os.path.dirname(__file__), "steps_schema.yml")
        with open(steps_schema_path, "r") as step_schema_file:
            steps_schema = yaml.safe_load(step_schema_file.read())
    except FileNotFoundError as e:
        raise FileNotFoundError("Steps schema file not found") from e

    try:
        with open(step_file_path, "r") as step_file:
            steps = yaml.safe_load(step_file.read())
            validate(instance=steps, schema=steps_schema)

    except ValidationError as e:
        raise ValidationError(
            f"Steps file has the following syntax errors: {e.message} "
        ) from e
    except FileNotFoundError as e:
        raise FileNotFoundError(f"Steps file {step_file_path} not found") from e

    print("[bold green]--Successfully Validated Steps File--[/bold green]")
    return steps


def parse_spec_file(steps: dict, spec_file_path: str) -> tuple:
    """Parse a valid OpenAPI document into a python dictionary

    Args:
        spec_file_path (str): Relative path to OpenAPI spec file

    Raises:
        OpenAPIValidationError: Invalid API Definition
        FileNotFoundError: Spec file not found
        UnsupportedSchemaError: Schema not supported

    Returns:
        tuple: Returns a tuple containing two dictionaries - the parsed API spec and
            a dictionary with all the openapi operations for quick reference
    """
    try:
        spec = materialize(RefDict(spec_file_path))
        validate_spec(spec)
    except OpenAPISpecValidatorError as e:
        raise OpenAPISpecValidatorError(
            f"API Spec file has the following syntax errors: {e} "
        ) from e
    except ValidationError as e:
        raise ValidationError(
            f"API Spec file has the following syntax errors: {e.message} "
        ) from e
    except FileNotFoundError as e:
        raise FileNotFoundError(f"API Spec file {spec_file_path} not found") from e

    # Get list of steps operationIds
    steps_operation_id_list = []
    for step in steps["steps"]:
        steps_operation_id_list.append(step["operation_id"])

    # Create operation schemas dict for easier parsing
    operation_schemas: dict = {}
    for path in spec["paths"]:
        for method in spec["paths"][path]:
            op_id = spec["paths"][path][method]["operationId"]
            if op_id in steps_operation_id_list:
                operation = spec["paths"][path][method]
                operation_schemas[op_id] = {}

                # Parse and Validate Request Bodies
                if "requestBody" in operation.keys():
                    # TODO: Support multiple content types if one of them is JSON
                    if len(operation["requestBody"]["content"]) > 1:
                        if (
                            "application/json"
                            not in operation["responses"][status_code]["content"]
                        ):
                            raise UnsupportedSchemaError(
                                f"There are too many request body content types for"
                                f" the operation: {op_id}"
                            )
                    if "application/json" in operation["requestBody"]["content"].keys():
                        operation_schemas[op_id]["requestBody"] = operation[
                            "requestBody"
                        ]["content"]["application/json"]["schema"]
                    elif (
                        "application/octet-stream"
                        in operation["requestBody"]["content"].keys()
                    ):
                        operation_schemas[op_id]["requestBody"] = operation[
                            "requestBody"
                        ]["content"]["application/octet-stream"]["schema"]
                    else:
                        raise UnsupportedSchemaError(
                            f"request body content type: "
                            f"{list(operation['requestBody']['content'].keys())[0]}"
                            f" unsupported for the operation: {op_id}"
                        )

                # Parse and Validate Response Bodies
                operation_schemas[op_id]["responses"] = {}
                for status_code in operation["responses"].keys():
                    if "content" in operation["responses"][status_code].keys():
                        # TODO: Support multiple content types if one of them is JSON
                        if len(operation["responses"][status_code]["content"]) > 1:
                            if (
                                "application/json"
                                not in operation["responses"][status_code]["content"]
                            ):
                                raise UnsupportedSchemaError(
                                    f"There are too many response body content types for"
                                    f" the operation: {op_id}"
                                )
                        if (
                            "application/json"
                            in operation["responses"][status_code]["content"].keys()
                        ):
                            operation_schemas[op_id]["responses"][
                                status_code
                            ] = operation["responses"][status_code]["content"][
                                "application/json"
                            ][
                                "schema"
                            ]
                        elif (
                            "application/octet-stream"
                            in operation["responses"][status_code]["content"].keys()
                        ):
                            operation_schemas[op_id]["responses"][
                                status_code
                            ] = operation["responses"][status_code]["content"][
                                "application/octet-stream"
                            ][
                                "schema"
                            ]
                        else:
                            raise UnsupportedSchemaError(
                                f"response body content type: "
                                f"{list(operation['responses'][status_code]['content'].keys())[0]}"
                                f" unsupported for the operation: {op_id}"
                            )
                    else:
                        operation_schemas[op_id]["responses"][status_code] = operation[
                            "responses"
                        ][status_code]

    print("[bold green]--Successfully Validated Spec File--[/bold green]")
    return spec, operation_schemas



import requests
from requests.auth import HTTPBasicAuth
from requests.structures import CaseInsensitiveDict
from typing import Union

from src.models.response import Response


class Request:
    def __init__(
        self,
        base_url: str,
        method: str,
        path: str,
        params: dict,
        headers: dict,
        global_auth: dict,
        auth: dict,
        body: Union[dict, str],
    ):
        self.base_url = base_url
        self.method = method
        self.path = path
        self.params = params
        self.headers = headers
        self.global_auth = global_auth
        self.auth = auth
        self.body = body

        if not self.auth:
            if not self.global_auth:
                self.auth = {"username": "", "password": ""}
            else:
                self.auth = self.global_auth

    def send(self):
        if isinstance(self.body, str):
            r = requests.request(
                url=self.base_url + self.path,
                method=self.method,
                params=self.params,
                headers=self.headers,
                # TODO: Beef up this logic to ensure a strict data type (binary/bytes?)
                data=self.body,
                auth=HTTPBasicAuth(self.auth["username"], self.auth["password"]),
            )
            return Response(
                status_code=r.status_code, headers=dict(r.headers), body=str(r.text)
            )
        else:
            r = requests.request(
                url=self.base_url + self.path,
                method=self.method,
                params=self.params,
                headers=self.headers,
                json=self.body,
                auth=HTTPBasicAuth(self.auth["username"], self.auth["password"]),
            )
            return Response(
                status_code=r.status_code, headers=dict(r.headers), body=str(r.text)
            )



from .preprocessing import evaluate
from .models import Step
from .verify import *

from rich import print, print_json
import json
from typing import Union


def make_requests(
    steps_data: dict, operation_schemas: dict, fail_fast: bool, verbose: bool
):

    base_url: str = steps_data["base_url"]

    # Set global auth if it exists in the step file
    global_auth: Union[dict, None] = None
    if "auth" in steps_data.keys():
        global_auth: dict = {}
        for key, value in steps_data["auth"].items():
            global_auth[key] = evaluate(value)
    else:
        global_auth = None

    # Get steps list
    steps_data: list = steps_data["steps"]

    # Create Global Steps List
    steps: dict = {}

    for s in steps_data:
        print(f"Step Name: {s['name']}")
        step = Step(s, steps)
        request = step.construct_request(base_url, global_auth)

        try:
            response_schema = operation_schemas[s["operation_id"]]["responses"][
                str(s["status_code"])
            ]
        except KeyError as e:
            print("[bold red]Response Validation Error[/bold red]")
            raise KeyError(
                f"Response code of {e} not found in responses for {s['operation_id']}"
            ) from e

        if "requestBody" in operation_schemas[s["operation_id"]].keys():
            verify_request_body(
                request.body, operation_schemas[s["operation_id"]]["requestBody"]
            )

        response = request.send()
        if verbose:
            if response.body:
                print("---Response---")
                print(f"Status Code: {response.status_code}")
                print_json(response.body)

        response_type = ""
        if (
            "type"
            in operation_schemas[s["operation_id"]]["responses"][
                str(s["status_code"])
            ].keys()
        ):
            response_type = operation_schemas[s["operation_id"]]["responses"][
                str(s["status_code"])
            ]["type"]

        if response_type == "object" or response_type == "array":
            try:
                response.body = json.loads(response.body)
            except json.decoder.JSONDecodeError as e:
                print("[bold red]Response Validation Error[/bold red]")
                raise Exception(
                    f"Response data is not valid JSON: {e}"
                ) from e

        verify_response(response, s["status_code"], response_schema)
        if verbose:
            print("[bold green]--Step Verified--[/bold green]")

        steps[s["name"]] = {}
        steps[s["name"]]["response"] = response



import pytest
import requests
from jsonschema import ValidationError

from src.parsing import parse_steps_file, parse_spec_file
from src.requests import make_requests


class TestRequests:
    """Class for basic unit testing of the requests.py module"""

    steps_file_path = "./tests/fixtures/valid/steps/person_api.yml"
    spec_file_path = "./tests/fixtures/valid/specs/person_api.yml"

    valid_test_steps = parse_steps_file(steps_file_path)
    valid_test_spec, operationSchemas = parse_spec_file(valid_test_steps, spec_file_path)

    def test_make_requests_with_valid_steps_and_spec(self, mocker):
        mock_response1 = requests.Response()
        mock_response1.status_code = 201
        mock_response1.headers = {}
        mock_response1.body = '{"id": 1}'
        mock_response1._content = b'{"id": 1}'
        mock_response1._text = mock_response1.content.decode('utf-8')

        mock_response2 = requests.Response()
        mock_response2.status_code = 200
        mock_response2.headers = {}
        mock_response2.body = '{"age": 29, "name": {"first": "test", "last": "test"}}'
        mock_response2._content = b'{"age": 29, "name": {"first": "test", "last": "test"}}'
        mock_response2._text = mock_response1.content.decode('utf-8')

        mock_responses = [mock_response1, mock_response2]
        mock_request = mocker.Mock(side_effect=mock_responses)
        mocker.patch('requests.request', new=mock_request)

        make_requests(self.valid_test_steps, self.operationSchemas, False, False)

    def test_make_requests_with_invalid_response(self, mocker):
        mock_response1 = requests.Response()
        mock_response1.status_code = 201
        mock_response1.headers = {}
        mock_response1.body = '{"id": 1}'
        mock_response1._content = b'{"id": 1}'
        mock_response1._text = mock_response1.content.decode('utf-8')

        mock_response2 = requests.Response()
        mock_response2.status_code = 200
        mock_response2.headers = {}
        # required property age is missing
        mock_response2.body = '{"name": {"first": "test", "last": "test"}}'
        mock_response2._content = b'{"name": {"first": "test", "last": "test"}}'
        mock_response2._text = mock_response1.content.decode('utf-8')

        mock_responses = [mock_response1, mock_response2]
        mock_request = mocker.Mock(side_effect=mock_responses)
        mocker.patch('requests.request', new=mock_request)

        with pytest.raises(ValidationError):
            make_requests(self.valid_test_steps, self.operationSchemas, False, False)



import pytest
from hamcrest import assert_that, is_, has_items, instance_of

from src.parsing import *
from src.preprocessing import *


class TestPreProcessing:
    """Class for basic unit testing of the preprocessing.py module"""

    steps_file_path = "./tests/fixtures/valid/steps/person_api.yml"
    spec_file_path = "./tests/fixtures/valid/specs/person_api.yml"

    steps = parse_steps_file(steps_file_path)
    spec, operation_schemas = parse_spec_file(steps, spec_file_path)

    def test_get_operation_coverage(self):
        step_coverage, spec_coverage = get_operation_coverage(self.steps, self.spec)
        assert_that(step_coverage == spec_coverage)

    def test_check_operation_coverage(self):
        check_operation_coverage(self.steps, self.spec)

    def test_evaluate_expression(self):
        evaluated_empty_expression = evaluate(None)
        assert_that(evaluated_empty_expression, is_(None))

        evaluated_env_var_expression = evaluate("${{ env.PYPONY_RELEASE_VERSION }}")
        assert_that(evaluated_env_var_expression, is_("0.0.0"))

        # TODO: More tests for evaluate()
        # evaluated_steps_expression = evaluate("${{ steps.someStep }}", {})



import pytest
from hamcrest import assert_that, is_, has_items, instance_of

from src.verify import *
from src.models import Response


class TestVerify:
    """Class for basic unit testing of the verify.py module"""

    schema = {
        "type": "object",
        "properties": {"name": {"type": "string"}, "age": {"type": "integer"}},
    }

    def test_verify_request_body_with_valid_request_body(self):
        request_body = {"name": "John Doe", "age": 30}
        assert_that(verify_request_body(request_body, self.schema), is_(None))

    def test_verify_request_body_with_invalid_request_body(self):
        request_body = {"name": "John Doe", "age": "30"}
        schema = {
            "type": "object",
            "properties": {"name": {"type": "string"}, "age": {"type": "integer"}},
        }
        with pytest.raises(ValidationError):
            verify_request_body(request_body, self.schema)

    def test_verify_response_with_valid_response(self):
        status_code = 200
        body = {"name": "John Doe", "age": 30}
        response = Response(status_code, {}, body)
        schema = {
            "type": "object",
            "properties": {"name": {"type": "string"}, "age": {"type": "integer"}},
        }

        assert_that(verify_response(response, status_code, self.schema), is_(None))

    def test_verify_response_with_invalid_response(self):
        status_code = 200
        data = {"name": "John Doe", "age": "30"}
        response = Response(status_code, {}, data)
        schema = {
            "type": "object",
            "properties": {"name": {"type": "string"}, "age": {"type": "integer"}},
        }

        with pytest.raises(ValidationError):
            verify_response(response, status_code, self.schema)



from jsonschema import validate, ValidationError
from rich import print
from typing import Union

from .models import Response


def verify_request_body(request_body: Union[dict, str], schema: dict):
    """Verify that the request body matches the schema provided in the step file"""

    try:
        validate(instance=request_body, schema=schema)
    except ValidationError as e:
        print("[bold red]--Request Validation Failed--[/bold red]")
        raise ValidationError("There was an issue with your request body.") from e
    return


def verify_response(response: Response, status_code: int, schema: dict):
    """Verify that the response matches the schema provided in the step file"""

    if response.status_code != status_code:
        print("[bold red]--Response Validation Failed--[/bold red]")
        raise ValidationError(
            "HTTP Status Code does not match the expected value from the step file."
        )
    else:
        try:
            validate(instance=response.body, schema=schema)
        except ValidationError as e:
            print("[bold red]--Response Validation Failed--[/bold red]")
            print(e)
            raise ValidationError("There was an issue with the response from the API.")
    return



import sys
import traceback

import click

from src.validate import validate


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "-st", "--step_file", required=True, type=click.STRING, envvar="INPUT_STEP_FILE"
)
@click.option(
    "-sp", "--spec_file", required=True, type=click.STRING, envvar="INPUT_SPEC_FILE"
)
@click.option("-ff", "--fail-fast", is_flag=True)
@click.option("-v", "--verbose", is_flag=True)
@click.version_option()
@click.help_option()
def main(step_file, spec_file, fail_fast, verbose):
    try:
        validate(step_file, spec_file, fail_fast, verbose)
    except BaseException as e:
        if verbose:
            print(traceback.format_exc())
        else:
            print(str(e))

        sys.exit(1)


main()



import pytest
from hamcrest import assert_that, is_, has_items, instance_of

from src.validate import *


class TestValidate:
    """Class for basic unit testing of the validate.py module"""

    steps_file_path = "./tests/fixtures/valid/steps/person_api.yml"
    spec_file_path = "./tests/fixtures/valid/specs/person_api.yml"

    @pytest.mark.skip(reason="Need to test against a mock API")
    def test_validate(self):
        validate(self.steps_file_path, self.spec_file_path, False, False)



from .parsing import parse_steps_file, parse_spec_file
from .preprocessing import check_operation_coverage
from .requests import make_requests

# from .verify import verify_request


def validate(
    step_file_path: str,
    spec_file_path: str,
    fail_fast: bool = False,
    verbose: bool = False,
):

    # convert step and spec into usable dictionaries
    print("--Validating Steps--")
    steps = parse_steps_file(step_file_path)

    print("--Validating Spec--")
    spec, operation_schemas = parse_spec_file(steps, spec_file_path)

    # Validate that desired coverage threshold is met (if present)
    check_operation_coverage(steps, spec)

    print("--Making Requests--")
    make_requests(steps, operation_schemas, fail_fast, verbose)
